/*
 *	D I S P L A Y . C
 *
 *	purpose:
 *		The functions in this file handle redisplay. There are two halves, the
 *		ones that Update the virtual display screen, and the ones that make the
 *		physical display screen the same as the virtual display screen. These
 *		functions use hints that are left in the windows by the commands.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <alloc.h>
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "version.h"
#include "protos.h"

#define VFCHG	0x0001		/* changed flag */
#define VFEXT	0x0002		/* extended (beyond column 80) */
#define VFCOL	0x0004		/* color change requested */

struct video
{
	int 	v_flag; 		/* flags */
	int 	v_fcolor;		/* current foreground color */
	int 	v_bcolor;		/* current background color */
	int 	v_rfcolor;		/* requested foreground color */
	int 	v_rbcolor;		/* requested background color */
	char	v_text[1];		/* screen data */
};
typedef	struct video	VIDEO;

/*
 * local function prototypes
 */

static void near	mlputc(int);
static void near	modeline(struct window *);
static void near	reframe(struct window *);
static void near	updall(struct window *);
static void near	updateline(int, struct video *);
static void near	upddex(void);
static void near	updext(void);
static void near	updgar(void);
static void near	updone(struct window *);
static void near	updpos(void);
static void near	updupd(void);
static void near	vteeol(void);
static void near	vtmove(int, int);
static void near	vtputc(int);

/*
 * local (module global) variable declarations
 */

static VIDEO	**vscreen;			/* virtual screen */

/**
 * Initialize the data structures used by the display code. The edge vectors
 * used to access the screens are set up. The terminal structure is allocated.
 * All the other things get initialized at compile time. The original window
 * has "WFCHG" set, so that it will get completely redrawn on the first call
 * to Update.
 */
void VtInit()
{
  int	i;
  VIDEO	*vp;

	if (!PcOpen())				/* open the screen */
		Quit(TRUE, FAILURE);

	vscreen = (VIDEO **)malloc(term.t_mrow * sizeof(VIDEO *));

	if (vscreen == (VIDEO **)NULLCPTR)
		Quit(TRUE, FAILURE);

	for (i = 0; i < term.t_mrow; ++i)
	{
		vp = (VIDEO *)malloc(sizeof(VIDEO) + term.t_mcol);

		if (vp == NULLVPTR)
			Quit(TRUE, FAILURE);

		vp->v_flag = 0;
		vp->v_rfcolor = 15;		/* color = "IWHI" */
		vp->v_rbcolor = 1;		/* color = "blu" */
		vscreen[i] = vp;
	}
}

/**
 * Clean up the virtual terminal system, in anticipation of a return to the
 * operating system.
 */
void VtTidy()
{
	PcClose();
}

/**
 * Set the virtual cursor to the specified row and column on the virtual
 * screen. There is no checking for nonsense values.
 */
static void near vtmove(row, col)
  int	row, col;
{
	vtrow = row;
	vtcol = col;
}

/**
 * Write a character to the virtual screen. Update the virtual row and column.
 * If we are not yet on left edge, don't print it. If the line is too long,
 * put a "$" in the last column. This routine only puts printing characters
 * into the virtual terminal buffers. Only column overflow is checked.
 */
static void near vtputc(c)
  int	c;
{
  VIDEO	*vp; 	/* ptr to line being updated */

	vp = vscreen[vtrow];

	if (c == '\t')
	{
		do
		{
			vtputc(' ');
		} while (((vtcol + taboff) % tabsize) != 0);
	}
	else if (vtcol >= term.t_ncol)
	{
		++vtcol;
		vp->v_text[term.t_ncol - 1] = '$';
	}
	else if (c < 0x20 || c == 0x7F)
	{
		vtputc('^');
		vtputc(c ^ 0x40);		/* e.g., show 0x07 ascii as ^G */
	}
	else
	{
		if (vtcol >= 0)
			vp->v_text[vtcol] = (char)c;
		++vtcol;
	}
}

/**
 * Erase from the end of the software cursor to the end of the line on which
 * the software cursor is located.
 */
static void near vteeol()
{
  VIDEO	*vp;

	vp = vscreen[vtrow];
	while (vtcol < term.t_ncol)
		vp->v_text[vtcol++] = ' ';
}

/**
 * User routine to force a complete screen Update
 */
int UpdateScr()
{
	return(Update(TRUE));
}

/**
 * Make sure that the display is right. This is a three part process. First,
 * scan through all of the windows looking for dirty ones. Check the framing,
 * and Refresh the screen. Second, make sure that "currow" and "curcol" are
 * correct for the current window. Third, make the virtual and physical
 * screens the same.
 */
int Update(force)
  int		force;		/* force Update past type ahead? */
{
  WINDOW	*wp;

	if (force == FALSE && kbdmode == PLAY)
		return(TRUE);

	/* Update any windows that need refreshing */
	wp = wheadp;
	while (wp != NULLWPTR)
	{
		if (wp->w_flag)
		{
			/* if the window has changed, service it */
			reframe(wp);							/* check the framing */
			if ((wp->w_flag & ~WFMODE) == WFEDIT)
				updone(wp); 						/* Update EDITed line */
			else if (wp->w_flag & ~WFMOVE)
				updall(wp); 						/* Update all lines */
			if (wp->w_flag & WFMODE)
				modeline(wp);						/* Update modeline */
			wp->w_flag = 0;
			wp->w_force = 0;
		}
		/* on to the next window */
		wp = wp->w_wndp;
	}

	/* recalc the current hardware cursor location */
	updpos();

	/* Update the cursor and flush the buffers */
	MoveCursor(currow, curcol - lbound);

	/* check for lines to de-extend */
	upddex();

	/* if screen is garbage, re-plot it */
	if (sgarbf != FALSE)
		updgar();

	/* Update the virtual screen to the physical screen */
	updupd();

	/* Update the cursor and flush the buffers */
	MoveCursor(currow, curcol - lbound);
	return(TRUE);
}

/**
 * Check to see if the cursor is in the window and re-frame it
 */
static void near reframe(wp)
  WINDOW	*wp;
{
  LINE		*lp;
  int		i;
  int		nlines;

	nlines = wp->w_ntrows;

	/* if not a requested reframe, check for a needed one */
	if ((wp->w_flag & WFFORCE) == 0)
	{
		lp = wp->w_linep;
		for (i = 0; i < nlines; i++)
		{
			/* if the line is in the window, no reframe */
			if (lp == wp->w_dotp)
				return;

			/* if we are at the end of the file, reframe */
			if (lp == wp->w_bufp->b_linep)
				break;

			/* on to the next line */
			lp = lforw(lp);
		}
	}

	/* reaching here, we need a window Refresh */
	i = wp->w_force;

	/* how far back to reframe? */
	if (i > 0)					/* only one screen worth of lines max */
	{
		if (--i >= nlines)
			i = nlines - 1;
	}
	else if (i < 0) 			/* negative Update???? */
	{
		i += nlines;
		if (i < 0)
			i = 0;
	}
	else
		i = nlines / 2;

	/* backup to new line at top of window */
	lp = wp->w_dotp;
	while (i != 0 && lback(lp) != wp->w_bufp->b_linep)
	{
		--i;
		lp = lback(lp);
	}

	/* and reset the current line at top of window */
	wp->w_linep = lp;
	wp->w_flag |= WFHARD;
	wp->w_flag &= ~WFFORCE;
}

/**
 * Update the current line to the virtual screen
 */
static void near updone(wp)
  WINDOW	*wp; 	/* window to Update current line in */
{
  LINE		*lp; 	/* line to Update */
  int		sline;	/* physical screen line to Update */
  int		i;

	/* search down the line we want */
	lp = wp->w_linep;
	sline = wp->w_toprow;
	while (lp != wp->w_dotp)
	{
		++sline;
		lp = lforw(lp);
	}

	/* and Update the virtual line */
	vscreen[sline]->v_flag |= VFCHG;
	taboff = wp->w_fcol;
	vtmove(sline, -taboff);
	for (i = 0; i < llength(lp); ++i)
		vtputc(lgetc(lp, i));
	vscreen[sline]->v_rfcolor = wp->w_fcolor;
	vscreen[sline]->v_rbcolor = wp->w_bcolor;
	vteeol();
	taboff = 0;
}

/**
 * Update all the lines in a window on the virtual screen
 */
static void near updall(wp)
  WINDOW	*wp; 	/* window to Update lines in */
{
  LINE		*lp; 	/* line to Update */
  int		sline;	/* physical screen line to Update */
  int		i;

	/* search down the lines, updating them */
	lp = wp->w_linep;
	sline = wp->w_toprow;
	taboff = wp->w_fcol;
	while (sline < wp->w_toprow + wp->w_ntrows)
	{
		/* and Update the virtual line */
		vscreen[sline]->v_flag |= VFCHG;
		vtmove(sline, -taboff);
		if (lp != wp->w_bufp->b_linep)
		{
			/* if we are not at the end */
			for (i = 0; i < llength(lp); ++i)
				vtputc(lgetc(lp, i));
			lp = lforw(lp);
		}

		/* make sure we are on screen */
		if (vtcol < 0)
			vtcol = 0;

		/* on to the next one */
		vscreen[sline]->v_rfcolor = wp->w_fcolor;
		vscreen[sline]->v_rbcolor = wp->w_bcolor;
		vteeol();
		++sline;
	}
	taboff = 0;
}

/**
 * Update the position of the hardware cursor and handle extended lines.
 * This is the only Update for simple moves.
 */
static void near updpos()
{
  LINE	*lp;
  int	c;
  int	i;

	/* find the current row */
	lp = curwp->w_linep;
	currow = curwp->w_toprow;
	while (lp != curwp->w_dotp)
	{
		++currow;
		lp = lforw(lp);
	}

	/* find the current column */
	curcol = 0;
	i = 0;
	while (i < curwp->w_doto)
	{
		c = lgetc(lp, i++);
		if (c == '\t')
			curcol += -(curcol % tabsize) + (tabsize - 1);
		else if (c < 0x20 || c == 0x7f)
			++curcol;
		++curcol;
	}

	/* adjust by the current first column position */
	curcol -= curwp->w_fcol;

	/* make sure it is not off the left side of the screen */
	while (curcol < 0)
	{
		if (curwp->w_fcol >= hjump)
		{
			curcol += hjump;
			curwp->w_fcol -= hjump;
		}
		else
		{
			curcol += curwp->w_fcol;
			curwp->w_fcol = 0;
		}
		curwp->w_flag |= WFHARD | WFMODE;
	}

	/* if horizontall scrolling is enabled, shift if needed */
	if (hscroll)
	{
		while (curcol >= term.t_ncol - 1)
		{
			curcol -= hjump;
			curwp->w_fcol += hjump;
			curwp->w_flag |= WFHARD | WFMODE;
		}
	}
	else
	{
		/* if extended, flag so and Update the virtual line image */
		if (curcol >= term.t_ncol - 1)
		{
			vscreen[currow]->v_flag |= (VFEXT | VFCHG);
			updext();
		}
		else
			lbound = 0;
	}
	/* Update the current window if we have to move it around */
	if (curwp->w_flag & WFHARD)
		updall(curwp);
	if (curwp->w_flag & WFMODE)
		modeline(curwp);
	curwp->w_flag = 0;
}

/**
 * Deextend any line that deserves it
 */
static void near upddex()
{
  WINDOW	*wp;
  LINE		*lp;
  int		i;
  int		j;

	wp = wheadp;

	while (wp != NULLWPTR)
	{
		lp = wp->w_linep;
		i = wp->w_toprow;

		while (i < wp->w_toprow + wp->w_ntrows)
		{
			if (vscreen[i]->v_flag & VFEXT)
			{
				if (wp != curwp || lp != wp->w_dotp || curcol < term.t_ncol - 1)
				{
					taboff = wp->w_fcol;
					vtmove(i, -taboff);
					for (j = 0; j < llength(lp); ++j)
						vtputc(lgetc(lp, j));
					vteeol();
					taboff = 0;

					/* this line no longer is extended */
					vscreen[i]->v_flag &= ~VFEXT;
					vscreen[i]->v_flag |= VFCHG;
				}
			}
			lp = lforw(lp);
			++i;
		}
		/* and onward to the next window */
		wp = wp->w_wndp;
	}
}

/**
 * If the screen is garbage, clear the physical screen and the virtual screen
 * and force a full Update
 */
static void near updgar()
{
  int	i;

	for (i = 0; i < term.t_nrow; ++i)
	{
		vscreen[i]->v_flag |= VFCHG;
		vscreen[i]->v_fcolor = gfcolor;
		vscreen[i]->v_bcolor = gbcolor;
	}

	MoveCursor(0, 0);	/* erase the screen */
	PcEeop();
	sgarbf = FALSE;
	mpresf = FALSE;
	MlErase();
}

/**
 * Update the physical screen from the virtual screen
 */
static void near updupd()
{
  VIDEO	*vp1;
  int	i;

	for (i = 0; i < term.t_nrow; ++i)
	{
		vp1 = vscreen[i];

		/* for each line that needs to be updated */
		if ((vp1->v_flag & VFCHG) != 0)
			updateline(i, vp1);
	}
}

/**
 * Update the extended line which the cursor is currently on at a column
 * greater than the terminal width. The line will be scrolled right or left
 * to let the user see where the cursor is.
 */
static void near updext()
{
  int	rcursor;		/* real cursor location */
  LINE	*lp; 			/* pointer to current line */
  int	i;				/* index into line */

	/* calculate what column the real cursor will end up in */
	rcursor = ((curcol - term.t_ncol) % term.t_scrsiz) + term.t_margin;
	lbound = curcol - rcursor + 1;
	taboff = lbound + curwp->w_fcol;

	/* scan through the line outputting characters to the virtual screen
	 * once we reach the left edge
	 */
	vtmove(currow, -taboff);			/* start scanning offscreen */
	lp = curwp->w_dotp; 				/* line to output */
	for (i = 0; i < llength(lp); ++i)	/* until the end-of-line */
		vtputc(lgetc(lp, i));

	/* truncate the virtual line, restore Tab offset */
	vteeol();
	taboff = 0;

	/* and put a '$' in column 1 */
	vscreen[currow]->v_text[0] = '$';
}

/**
 * Update a single line. This does not know how to use insert or delete
 * character sequences. Update the physical row and column variables.
 */
static void near updateline(row, vp1)
  int			row;	/* row of screen to Update */
  struct video	*vp1;	/* virtual screen image */
{
	PcWrite(row, vp1->v_text, vp1->v_rfcolor, vp1->v_rbcolor);
	vp1->v_fcolor = vp1->v_rfcolor;
	vp1->v_bcolor = vp1->v_rbcolor;
	vp1->v_flag &= ~(VFCHG | VFCOL);	/* flag this line as changed */
}

/**
 * Redisplay the mode line for the window pointed to by the "wp". This is the
 * only routine that has any idea of how the modeline is formatted. You can
 * change the modeline format by hacking at this routine. Called by "Update"
 * any time there is a dirty window.
 */
static void near modeline(wp)
  WINDOW	*wp;
{
  char		*cp;
  int		c;
  int		n;				/* cursor position count */
  BUFFER	*bp;
  int		i;				/* loop index */
  int		lchar;			/* character to draw line in buffer with */
  char		tline[NLINE];	/* buffer for part of mode line */
  char		mline[2];		/* single char string bufr for mode codes */

	n = wp->w_toprow + wp->w_ntrows;		/* location */
	vscreen[n]->v_flag |= VFCHG | VFCOL;	/* redraw next time */

	vscreen[n]->v_rfcolor = 0;				/* black on */
	vscreen[n]->v_rbcolor = 7;				/* white */

	vtmove(n, 0);							/* seek to right line */
	if (wp == curwp)
		lchar = '';						/* '' if active buffer */
	else
		lchar = '';						/* '' if inactive buffer */

	bp = wp->w_bufp;
	if ((bp->b_flag & BFTRUNC) != 0)		/* '#' if truncated */
		vtputc('#');
	else
		vtputc(lchar);

	if ((bp->b_flag & BFCHG) != 0)			/* '*' if changed */
		vtputc('*');
	else
		vtputc(lchar);

	if ((bp->b_flag & BFNAROW) != 0)		/* '%' if narrowed */
		vtputc('%');
	else
		vtputc(lchar);

	n = 3;									/* '#' + '*' + '%' */

	strfmt(tline, " %s %s", PROGNAME, VERSION);

	/* display the modes */
	stracat(tline," (");
	for (i = 0; i < NUMMODES; i++)	/* add in the mode flags */
	{
		if (wp->w_bufp->b_mode & (1 << i))
		{
			strfmt(mline, "%c", modecode[i]);
			stracat(tline, mline);
		}
		else
			stracat(tline, ".");
	}
	stracat(tline,") ");

	cp = &tline[0];
	while ((c = *cp++) != 0)
	{
		vtputc(c);
		++n;
	}

	vtputc(lchar);
	vtputc(lchar);
	vtputc(' ');
	n += 3;
	cp = &bp->b_bname[0];					/* buffer name */
	while ((c = *cp++) != 0)
	{
		vtputc(c);
		++n;
	}

	vtputc(' ');
	vtputc(lchar);
	vtputc(lchar);
	n += 3;

	if (bp->b_fname[0] != 0)				/* file name */
	{
		cp = " file: ";
		while ((c = *cp++) != 0)
		{
			vtputc(c);
			++n;
		}

		cp = &bp->b_fname[0];
		while ((c = *cp++) != 0)
		{
			vtputc(c);
			++n;
		}

		vtputc(' ');
		++n;
	}

	while (n < term.t_ncol) 				/* pad to full width if necessary */
	{
		vtputc(lchar);
		++n;
	}
}

/**
 * Update all the mode lines
 */
void UpMode()
{
  WINDOW	*wp;

	wp = wheadp;
	while (wp != NULLWPTR)
	{
		wp->w_flag |= WFMODE;
		wp = wp->w_wndp;
	}
}

/**
 * Force hard updates on all windows
 */
void UpWind()
{
  WINDOW	*wp;

	wp = wheadp;
	while (wp != NULLWPTR)
	{
		wp->w_flag |= WFHARD | WFMODE;
		wp = wp->w_wndp;
	}
}

/**
 * Send a command to the terminal to move the hardware cursor to row "row"
 * and column "col". The row and column arguments are origin 0. Optimize out
 * random calls. Update "ttrow" and "ttcol".
 */
void MoveCursor(row, col)
  int	row, col;
{
	if (row != ttrow || col != ttcol)
	{
		ttrow = row;
		ttcol = col;
		PcMove(row, col);
	}
}

/**
 * Write a message into the message line. Keep track of the physical cursor
 * position.
 */
void MlWrite(fstr)
  char	*fstr;	/* formatted string for output */
{
  int	c;		/* current char in formatted string */

	/* if we are not currently echoing on the command line, abort this */
	if (discmd == FALSE)
	{
		MoveCursor(term.t_nrow, 0);
		return;
	}

	/* set up the proper colors for the command line */
	PcForeClr(3);				/* color = "CYA" */
	PcBackClr(1);				/* color = "blu" */

	MoveCursor(term.t_nrow, 0);
 	lstptr = &lstmsg[0];		/* setup to record message */
	while ((c = *fstr++) != 0)
	{
		mlputc(c);
		++ttcol;
	}

	PcEeol();					/* erase the remainder of the line */
	mpresf = TRUE;
	*lstptr = NULLCHR;			/* terminate lstmsg[] */
}

/**
 * Force a string out to the message line regardless of the current $discmd
 * setting. This is needed when $debug is TRUE and for the write-message and
 * clear-message-line commands
 */
void MlForce(s)
  char	*s;				/* string to force out */
{
  int	oldcmd; 		/* original command display flag */

	oldcmd = discmd;	/* save the discmd value */
	discmd = TRUE;		/* and turn display on */
	MlWrite(s); 		/* write the string out */
	discmd = oldcmd;	/* and restore the original setting */
}

/**
 * Erase the message line. This is a special routine because the message line
 * is not considered to be part of the virtual screen.
 */
void MlErase()
{
	MoveCursor(term.t_nrow, 0);
	if (discmd == FALSE)
		return;

	PcForeClr(3);			/* color = "CYA" */
	PcBackClr(1);			/* color = "blu" */

	PcEeol();				/* erase the line */
	mpresf = FALSE;
}

/**
 * Write out a string. Update the physical cursor position. This assumes that
 * the characters in the string all have width "1"; if this is not the case
 * things will get screwed up a little.
 */
void MlPutS(s)
  char	*s;
{
  int	c;

	while ((c = *s++) != 0)
	{
		mlputc(c);
		++ttcol;
	}
}

/**
 * Write a character to the modeline
 */
static void near mlputc(c)
  int	c;
{
	PcPutC(c);
	*lstptr++ = (char)c;
}
