/*
 *	I B M P C . C
 *
 *	purpose:
 *		The routines in this file go directly to the video RAM to do screen
 *		output. Supported monitor cards are MONO, CGA, EGA, VGA except that
 *		there is no flicker-free (snow removal) support for the CGA.
 *
 *	technical notes:
 *		IMPORTANT!!	The video type defines (CDCGA, CDV50 etc.) MUST match
 *					the order of the names listed in drvname[].
 *
 *		Note:	I am ignoring the 17th triplet in the .PAL file; I have not
 *				yet determined its exact relevence.  I think it has something
 *				to do with the overscan area, but the files generated by PRISM
 *				place some values in the third set which are greater than 63
 *				(intense white).  The overscan area is getting set to the color
 *				of the 1st triplet, which is the (default) black register.
 *				Modifying the upper left-most color in PRISM will set the
 *				overscan (border) color for CodEd.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <conio.h>
#include <dos.h>
#include <mem.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#define	TERMDEF 1			/* don't define "term" external */

#include "struct.h"
#include "def.h"
#include "protos.h"

#define NROW	50			/* maximum number of rows */
#define NCOL	80			/* maximum number of columns */
#define MARGIN	8			/* size of minimim margin */
#define SCRSIZ	64			/* scroll size for extended lines */
#define NPAUSE	256 		/* # times thru Update to pause */

#define M_SEG	0xb000		/* segment address, page 0, of mono video RAM */
#define C_SEG	0xb800		/* segment address, page 0, of color video RAM */

#define CDSENSE -1			/* detect the card type */
#define CDMNO	0			/* MONO - 25 line mode */
#define CDCGA	1			/* CGA - 25 line mode */
#define CDE25	2			/* EGA - 25 line mode */
#define CDE35	3			/* EGA - 35 line mode */
#define CDE43	4			/* EGA - 43 line mode */
#define CDV25	5			/* VGA - 25 line mode */
#define CDV28	6			/* VGA - 28 line mode */
#define CDV40	7			/* VGA - 40 line mode */
#define CDV50	8			/* VGA - 50 line mode */ 

#define NDRIVE	9			/* number of screen drivers */

struct rgbMap
{
	uchar	red;
	uchar	green;
	uchar	blue;
};
typedef struct rgbMap		RGB;

/*
 * local function prototypes
 */

static int near		scinit(int);
static void near	badmode(char *);
static void near	egaset(int);
static void near	getboard(void);
static void near	setvgareg(void);
static void near	vgaset(int, char *);

/*
 * local (module global) variable declarations
 */

static int	dtype = -1; 			/* current display type */
static char	*drvname[] = 			/* display type names */
{
	"MNO", "CGA", "E25", "E35", "E43", "V25", "V28", "V40", "V50"
};

static ushort	*scptr[NROW];		/* pointer to screen lines */
static ushort	v_seg;				/* segment address of video RAM */

static int 		mnoexist = FALSE;	/* is MONO card available? */
static int 		cgaexist = FALSE;	/* is CGA card available? */
static int 		egaexist = FALSE;	/* is EGA card available? */
static int 		vgaexist = FALSE;	/* is VGA card available? */
static int 		cfcolor = -1;		/* current foreground color */
static int 		cbcolor = -1;		/* current background color */

static RGB		rgb[17];			/* place to store the palette file */
static uchar	egaReg[17] =		/* default contents of EGA registers */
{
	0, 1, 2, 3, 4, 5, 20, 7, 56, 57, 58, 59, 60, 61, 62, 63, 0
};

static FILE		*fp;				/* file pointer */
static char		*cePal = "ce.pal";	/* CodEd palette file (VGA only) */
static char		*dosPal = "dos.pal";/* DOS palette file (VGA only) */

static union REGS	rg;				/* registers for DOS call */

/*
 * standard terminal parameters
 */

TERM term =
{
	NROW - 1,
	NROW - 1,
	NCOL,
	NCOL,
	MARGIN,
	SCRSIZ,
	NPAUSE
};

/**
 * Initialize the display system
 */
int PcOpen()
{
	if (!scinit(CDSENSE))
		return(FALSE);

	/* we are not sure of the initial position of the cursor */
	ttrow = 999;
	ttcol = 999;
	return(TRUE);
}

/**
 * Terminate the display system
 */
void PcClose()
{
	/* reset video adapter to normal text mode */
	switch (dtype)
	{
		case CDMNO:
			MNO25();
			break;
		case CDCGA:
			CGA25();
			break;
		case CDE25:
		case CDE35:
		case CDE43:
			egaset(CDE25);
			break;
		case CDV25:
		case CDV28:
		case CDV40:
		case CDV50:
			vgaset(CDV25, dosPal);
			break;
	}

	/* any code here executes just before we exit or shell from CodEd */

}

/**
 * Erase to the end of line
 */
void PcEeol()
{
  ushort	attr;		/* attribute byte mask to place in RAM */
  ushort	ec;			/* char to erase with */
  ushort	*lnptr;		/* pointer to the destination line */
  int		i;
  ushort	ccol;		/* current cursor:	column */
  ushort	crow;		/*					row */

	/* find the current cursor position */
	rg.h.ah = 3;			/* read cursor position function code */
	rg.h.bh = 0;			/* current video page */
	int86(0x10, &rg, &rg);
	ccol = (ushort)((uchar)rg.h.dl);	/* record current column */
	crow = (ushort)((uchar)rg.h.dh);	/* and row */

	/* build the attribute byte and setup the screen pointer */
	if (dtype != CDMNO)
		attr = (((cbcolor & 7) << 4) | (cfcolor & 15)) << 8;
	else
		attr = 0x0700;
	ec = SPACE | attr;
	lnptr = scptr[crow] + ccol;

	/* and send the string out */
	for (i = ccol * 2; i < term.t_ncol * 2; i += 2)
		*lnptr++ = ec;
}

/**
 * Erase to the end of page
 */
void PcEeop()
{
	rg.h.ah = 6;			/* scroll page up function code */
	rg.h.al = 0;			/* # lines to scroll (clear it) */
	rg.h.ch = 0;			/* upper left corner of scroll */
	rg.h.cl = 0;
	rg.h.dh = term.t_nrow;	/* lower right corner of scroll */
	rg.h.dl = term.t_ncol - 1;

	if (dtype != CDMNO)
		rg.h.bh = (char)((uchar)((gbcolor & 7) << 4) | (gfcolor & 15));
	else
		rg.h.bh = 0;
	
	int86(0x10, &rg, &rg);
}

/**
 * Change screen resolution
 */
int PcRez(res)
  char	*res;		/* resolution to change to */
{
  int	i;

	for (i = 0; i < NDRIVE; i++)
	{
		if (stracmp(res, drvname[i]) == 0)
			return(scinit(i));
	}
	MlWrite("Invalid display mode");
	PcPutC(BELL);
	return(FALSE);
}

/**
 * Set the current foreground color
 */
void PcForeClr(color)
  int	color;		/* color to set */
{
	cfcolor = color;
}

/**
 * Set the current background color
 */
void PcBackClr(color)
  int	color;		/* color to set */
{
	cbcolor = color;
}

/**
 * Write a line out
 */
void PcWrite(row, outstr, forg, bacg)
  int		row;		/* row of screen to place outstr on */
  char		*outstr; 	/* string to write out (must be term.t_ncol long) */
  int		forg;		/* foreground color of string to write */
  int		bacg;		/* background color */
{
  ushort	attr;		/* attribute byte mask to place in RAM */
  ushort	*lnptr;		/* pointer to the destination line */
  int		i;

	/* build the attribute byte and setup the screen pointer */
	attr = (((bacg & 7) << 4) | (forg & 15)) << 8;
	lnptr = scptr[row];

	/* and send the string out */
	for (i = 0; i < term.t_ncol; i++)
		*lnptr++ = (ushort)((uchar)(outstr[i] & 0xff)) | attr;
}

/**
 * Initialize display to specified mode
 */
static int near scinit(type)
  int	type;			/* type of adapter to init for */
{
  int	i;

	/* if asked, find out what display is connected */
	if (type == CDSENSE)
	{
		getboard();
		if (mnoexist == TRUE)
			type = CDMNO;
		else if (cgaexist == TRUE)
			type = CDCGA;
		else if (egaexist == TRUE)
			type = CDE25;
		else	/* vgaexist == TRUE */
			type = CDV25;
	}

	/* if we try to switch to MONO and there is none, don't */
	if (type == CDMNO && mnoexist != TRUE)
	{
		badmode("MONO");
		return(FALSE);
	}

	/* if we try to switch to CGA and there is none, don't */
	if (type == CDCGA && cgaexist != TRUE)
	{
		badmode("CGA");
		return(FALSE);
	}

	/* if we try to switch to EGA and there is none, don't */
	if ((type == CDE25 || type == CDE35 || type == CDE43) && egaexist != TRUE)
	{
		badmode("EGA");
		return(FALSE);
	}

	/* if we try to switch to VGA and there is none, don't */
	if ((type == CDV25 || type == CDV28 || type == CDV40 || type == CDV50) && vgaexist != TRUE)
	{
		badmode("VGA");
		return(FALSE);
	}

	/* and set up the various parameters as needed */
	v_seg = C_SEG;					/* default to color video RAM */
	switch (type)
	{
		case CDMNO:					/* monochrome - 25 line mode */
			v_seg = M_SEG;			/* point to monochrome video RAM */
			MNO25();
			NewSize(TRUE, 25);
			break;
		case CDCGA: 				/* CGA - 25 line mode */
			CGA25();
			NewSize(TRUE, 25);
			break;
		case CDE25: 				/* EGA - 25 line mode */
			egaset(type);
			NewSize(TRUE, 25);
			break;
		case CDE35: 				/* EGA - 35 line mode */
			egaset(type);
			NewSize(TRUE, 35);
			break;
		case CDE43: 				/* EGA - 43 line mode */
			egaset(type);
			NewSize(TRUE, 43);
			break;
		case CDV25: 				/* VGA - 25 line mode */
			vgaset(type, cePal);
			NewSize(TRUE, 25);
			break;
		case CDV28: 				/* VGA - 28 line mode */
			vgaset(type, cePal);
			NewSize(TRUE, 28);
			break;
		case CDV40:					/* VGA - 40 line mode */
			vgaset(type, cePal);
			NewSize(TRUE, 40);
			break;
		case CDV50:					/* VGA - 50 line mode */
			vgaset(type, cePal);
			NewSize(TRUE, 50);
			break;
	}

	/* reset the $sres environment variable */
	stracpy(sres, drvname[type]);
	dtype = type;

	/* initialize the screen pointer array */
	for (i = 0; i < NROW; i++)
		scptr[i] = MK_FP(v_seg, NCOL * i * 2);

	return(TRUE);
}

/**
 * Determine which type of display board is attached. Current known types are:
 *
 *		CDMNO		Monochrome - 25 line mode
 *		CDCGA		CGA - 25 line mode
 *		CDE25		EGA - 25 line mode
 *		CDE35		EGA - 35 line mode
 *		CDE43		EGA - 43 line mode
 *		CDV25		VGA - 25 line mode
 *		CDV28		VGA - 28 line mode
 *		CDV40		VGA - 40 line mode
 *		CDV50		VGA - 50 line mode
 *
 * Detect the current display adapter:
 *
 *	if	MONO	MNOexist = TRUE (all others FALSE)
 *		CGA 	CGAexist = TRUE (all others FALSE)
 *		EGA 	EGAexist = TRUE (all others FALSE)
 *		VGA 	VGAexist = TRUE (all others FALSE)
 */
static void near getboard()
{
	/* test for MONO */
	int86(0x11, &rg, &rg);
	if ((((rg.x.ax >> 4) & 3) == 3))
	{
		mnoexist = TRUE;
		cgaexist = FALSE;
		egaexist = FALSE;
		vgaexist = FALSE;
		return;
	}

	/* no MONO, so test for CGA, EGA or VGA */
	rg.x.ax = 0x1200;
	rg.x.bx = 0xff10;
	int86(0x10,&rg, &rg);			/* if EGA, VGA bh=0-1 and bl=0-3 */
	egaexist = !(rg.x.bx & 0xfefc); /* yes, it's EGA */
	if (egaexist)
	{
		/* Adapter says it's an EGA. We'll get the same response
		 * from a VGA, so try to tell the two apart.
		 */
		rg.x.ax = 0x1a00;			/* read display combination */
		int86(0x10, &rg, &rg);
		if (rg.h.al == 0x1a && (rg.h.bl == 7 || rg.h.bl == 8))
		{
			/* Function is supported and it's a PS/2 50, 60, 70, 80 with
			 * an analog display, so it's VGA (I hope!)
			 */
			cgaexist = FALSE;
			egaexist = FALSE;
			vgaexist = TRUE;
		}
		else
		{
			/* Either the BIOS function is not supported or it's something
			 * other than a VGA, so set it to be EGA.
			 */
			cgaexist = FALSE;
			vgaexist = FALSE;		/* egaexist is already TRUE */
		}
	}
	else							/* assume poor old CGA */
	{
		cgaexist = TRUE;
		vgaexist = FALSE;			/* egaexist is already FALSE */
	}
}

/**
 * Put the beast into the specified EGA mode
 */
static void near egaset(type)
  int	type;
{
	switch (type)
	{
		case CDE25:
			EGA25();		/* 8x14 char set */
			break;

		case CDE35:
			EGA35();		/* 8x10 char set and select alt. PrtSc */
			break;

		case CDE43:
			EGA43();		/* 8x8 char set and select alt. PrtSc */
			break;
	}
}

/**
 * Put the beast into the specified VGA mode
 */
static void near vgaset(type, palFile)
  int		type;
  char		*palFile;
{
  char		*ufn;
  int		i;

	switch (type)
	{
		case CDV25:
			VGA25();		/* 8x16 char set */
			break;

		case CDV28:
			VGA28();		/* 8x14 char set/400 scan lines/select alt. PrtSc */
			break;

		case CDV40:
			VGA40();		/* 8x10 char set and select alt. PrtSc */
			break;

		case CDV50:
			VGA50();		/* 8x8 char set and select alt. PrtSc */
			break;
	}

	if ((ufn = FileLook(palFile, TRUE)) == NULLCPTR)
		return;
	if ((fp = fopen(ufn, "rb")) == (FILE *)NULLCPTR)
		return;
	for (i = 0; i < 17; i++)
	{
		if (feof(fp))		/* just quit, without calling setvgareg() */
		{
			fclose(fp);
			return;
		}
		rgb[i].red = (uchar)fgetc(fp);
		rgb[i].green = (uchar)fgetc(fp);
		rgb[i].blue = (uchar)fgetc(fp);
	}
	fclose(fp);
	setvgareg();
}

/**
 * Set the 16 currently active VGA registers to the colors specified in
 * the palette array.  Note the extra level of indirection here - using
 * the default contents of the 16 (0 -> 15) EGA registers as a pointer to
 * the desired VGA register, rather than as the VGA register number itself.
 */
static void near setvgareg()
{
  int	i;

	for (i = 0; i < 16; i++)
	{
		rg.h.ah = 0x10;
		rg.h.al = 0x10;
		rg.h.bh = 0x00;
		rg.h.bl = egaReg[i];
		rg.h.dh = rgb[i].red;
		rg.h.ch = rgb[i].green;
		rg.h.cl = rgb[i].blue;
		int86(0x10, &rg, &rg);
	}
}

static void near badmode(mode)
  char	*mode;
{
  char	buf[NSTRING];

	strcpycat(buf, mode, " does not exist");
	sgarbf = TRUE;								/* screen is garbage */
	Update(TRUE);								/* so repaint it */
	MlWrite(buf);
	PcPutC(BELL);
}
