/*
 *	S T R U C T . H
 *
 *	purpose:
 *		Structure and preprocessor definitions for CodEd.
 */

/* configuration options */

#define BINARY	1		/* use FUNC, EV and UF Binary search code */
#define BMARKS	1		/* include bookmarks */
#define MAGIC	1 		/* include regular expression matching */

/* define some NULLs */

#define NULLCPTR	((char *)0)			/* same as PforCe NULLPTR */

#define NULLBPTR	(BUFFER *)NULLCPTR
#define NULLKPTR	(KILL *)NULLCPTR
#define NULLLPTR	(LINE *)NULLCPTR
#define NULLSPTR	(short *)NULLCPTR
#define NULLVPTR	(VIDEO *)NULLCPTR
#define NULLWPTR	(WINDOW *)NULLCPTR
#define NULLWHPTR	(WHBLOCK *)NULLCPTR

#define NULLFPTR	(int((*)()))NULLCPTR

#define NULLCHR		'\0'				/* string null terminator */

/* internal constants */

#define NBINDS	256 	/* max # of bound keys */
#define NFILEN	80		/* # of bytes, file name */
#define NBUFN	16		/* # of bytes, buffer name */
#define NLINE	256 	/* # of bytes, input line */
#define NSTRING 128 	/* # of bytes, string buffers */
#define NKBDM	256 	/* # of key strokes, keyboard macro */
#define NPAT	128 	/* # of bytes, pattern */
#define NCOLORS 16		/* number of supported colors */
#define KBLOCK	256		/* sizeof kill buffer chunks (MUST BE MULTIPLE OF 16) */
#define NBLOCK	16		/* sizeof line block chunks (MUST BE MULTIPLE OF 16) */
#define NVSIZE	10		/* max # of chars in a variable name */

#if BMARKS
#define NBMARKS	10		/* number of bookmarks */
#endif

/* character class flags (see kdef.h for key definitions) */

#define CTRL	0x0100	/* control flag */
#define META	0x0200	/* meta flag */
#define CTLX	0x0400	/* Ctrl-X flag */
#define SPEC	0x0800	/* special key (function and alt keys) */

#ifdef	FALSE
#undef	FALSE
#endif
#ifdef	TRUE
#undef	TRUE
#endif

#define FALSE	0		/* false, no, bad, etc. */
#define TRUE	1		/* true, yes, good, etc. */
#define ABORT	2		/* death, Alt-A, abort, etc. */

#define STOP	0		/* keyboard macro not in use */
#define PLAY	1		/* playing */
#define RECORD	2		/* recording */

/* directive definitions */

#define DIF 		0
#define DELSE		1
#define DENDIF		2
#define DGOTO		3
#define DRETURN 	4
#define DENDM		5
#define DWHILE		6
#define DENDWHILE	7
#define DBREAK		8
#define DFORCE		9

#define NUMDIRS 	10	/* number of directives defined */

/* toggleable values for the search Scanner routines */

#define PTBEG	0		/* leave "dot" at beginning of search string */
#define PTEND	1		/* leave "dot" at end of search string */
#define FORWARD 0		/* forward direction */
#define REVERSE 1		/* reverse direction */

/* file I/O return codes */

#define FIOSUC	0		/* success */
#define FIOFNF	1		/* file not found */
#define FIOEOF	2		/* end of file */
#define FIOERR	3		/* error */
#define FIOMEM	4		/* out of memory */
#define FIOFUN	5		/* end of file/bad line */

#define CFLMOV	0x0001	/* last command was a line move (ForwLine/BackLine) */
#define CFKILL	0x0002	/* last command was a kill */

#define SPACE	0x20	/* space character */

#define INTWIDTH		sizeof(int) * 3

/* macro argument Token types */

#define TKNUL	0		/* end-of-string */
#define TKARG	1		/* interactive argument */
#define TKBUF	2		/* buffer argument */
#define TKVAR	3		/* user variables */
#define TKENV	4		/* environment variables */
#define TKDIR	5		/* directive */
#define TKFUN	6		/* function */
#define TKLBL	7		/* line label */
#define TKLIT	8		/* numeric literal */
#define TKSTR	9		/* quoted string literal */
#define TKCMD	10		/* command name */

/* There is a window structure allocated for every active display window. The
 * windows are kept in a big list, in top to bottom screen order, with the
 * listhead at "wheadp". Each window contains its own values of dot and mark.
 * The flag field contains some bits that are set by commands to guide
 * redisplay. Although this is a bit of a compromise in terms of decoupling,
 * the full blown redisplay is just too expensive to run for every input
 * character.
 */

struct window
{
	struct window	*w_wndp; 			/* next window */
	struct buffer	*w_bufp; 			/* buffer displayed in window */
	struct line		*w_linep;			/* top line in the window */
	struct line 	*w_dotp; 			/* line containing "dot" */
	short			w_doto; 			/* byte offset for "dot" */
	struct line		*w_markp;			/* line containing "mark" */
	short			w_marko;			/* byte offset for "mark" */
	struct line		*w_anchorp;			/* line containing "anchor" */
	short			w_anchoro;			/* byte offset for "anchor" */
#if BMARKS
	struct line		*w_bmarkp[NBMARKS];	/* line containing "bookmark x" */
	short			w_bmarko[NBMARKS];	/* byte offset for "bookmark x" */
#endif
	char			w_toprow;			/* origin 0 top row of window */
	char			w_ntrows;			/* # of rows of text in window */
	char			w_force;			/* if NZ, forcing row */
	char			w_flag; 			/* flags */
	char			w_fcolor;			/* current foreground color */
	char			w_bcolor;			/* current background color */
	int				w_fcol;				/* first column displayed */
};
typedef struct window	WINDOW;

#define WFFORCE 0x01			/* window needs forced reframe */
#define WFMOVE	0x02			/* movement from line to line */
#define WFEDIT	0x04			/* editing within a line */
#define WFHARD	0x08			/* better to a full display */
#define WFMODE	0x10			/* Update mode line */
#define WFCOLR	0x20			/* needs a color change */

/* Text is kept in buffers. A buffer header, described below, exists for every
 * buffer in the system. The buffers are kept in a big list, so that commands
 * that search for a buffer by name can find the buffer header. There is a
 * safe store for the dot and mark in the header, but this is only valid if
 * the buffer is not being displayed (that is, if "b_nwnd" is 0). The text for
 * the buffer is kept in a circularly linked list of lines, with a pointer to
 * the header line in "b_linep". Buffers may be "Inactive" which means the
 * files associated with them have not been read in yet. These get read in at
 * "UseBuffer" time.
 */

struct buffer
{
	struct buffer	*b_fp;	 			/* next buffer */
	struct buffer	*b_bp;	 			/* previous buffer */
	struct line		*b_dotp; 			/* line containing "dot" */
	short			b_doto; 			/* byte offset for "dot" */
	struct line		*b_markp;			/* line containing "mark" */
	short			b_marko;			/* byte offset for "mark" */
	struct line		*b_anchorp;			/* line containing "anchor" */
	short			b_anchoro;			/* byte offset for "anchor" */
#if BMARKS
	struct line		*b_bmarkp[NBMARKS];	/* line containing "bookmark x" */
	short			b_bmarko[NBMARKS];	/* byte offset for "bookmark x" */
#endif
	int				b_fcol;				/* first col to display */
	struct line		*b_linep;			/* link to the header LINE */
	struct line		*b_topline;			/* link to top line of narrowed text */
	struct line		*b_botline;			/* link to bot line of narrowed text */
	char			b_active;			/* window activated flag */
	char			b_nwnd; 			/* count of windows on buffer */
	char			b_flag; 			/* flags */
	int 			b_mode; 			/* editor mode of this buffer */
	char			b_fname[NFILEN];	/* file name */
	char			b_bname[NBUFN]; 	/* buffer name */
};
typedef struct buffer	BUFFER;

#define BFINVS	0x01			/* internal invisible buffer */
#define BFCHG	0x02			/* changed since last write */
#define BFTRUNC 0x04			/* buffer was truncated when read */
#define BFNAROW	0x08			/* buffer has been narrowed	*/

/* mode flags */

#define MDCMODE	0x0001			/* C indentation and fence match */
#define MDEXACT 0x0002			/* exact matching for searches */
#define MDVIEW	0x0004			/* read-only buffer */
#define MDOVER	0x0008			/* overwrite mode */
#define MDMAGIC 0x0010			/* regular expressions in search */

#define NUMMODES		5		/* number of defined mode flags */

/* The starting position of a region, and the size of the region in
 * characters, is kept in a region structure.
 */

struct region
{
	struct line	*r_linep;		/* origin LINE address */
	short		r_offset;		/* origin LINE offset */
	long		r_size; 		/* length in characters */
};
typedef struct region	REGION;

/* All text is kept in circularly linked lists of "LINE" structures. These
 * begin at the header line (which is the blank line beyond the end of the
 * buffer). This line is pointed to by the "BUFFER". Each line contains the
 * number of bytes in the line (the "used" size), the size of the text array,
 * and the text. The end of line is not stored as a byte; it's implied. Future
 * additions will include Update hints, and a list of marks into the line.
 */

struct line
{
	struct line	*l_fp;			/* link to the next line */
	struct line	*l_bp;			/* link to the previous line */
	short		l_size; 		/* allocated size */
	short		l_used; 		/* used size */
	char		l_text[1];		/* a bunch of characters */
};
typedef struct line 	LINE;

#define lforw(lp)		((lp)->l_fp)
#define lback(lp)		((lp)->l_bp)
#define lgetc(lp, n)	((lp)->l_text[(n)]&0xFF)
#define lputc(lp, n, c) ((lp)->l_text[(n)]=(c))
#define llength(lp) 	((lp)->l_used)

/* The editor communicates with the display using a high level interface.
 * The "TERM" structure holds useful variables.
 */

struct term
{
	short	t_mrow; 			/* max number of rows allowable */
	short	t_nrow; 			/* current number of rows used */
	short	t_mcol; 			/* max number of columns */
	short	t_ncol; 			/* current number of columns */
	short	t_margin;			/* min margin for extended lines */
	short	t_scrsiz;			/* size of scroll region */
	int 	t_pause;			/* # times thru Update to pause */
};
typedef struct term 	TERM;

/* structure for the table of initial key bindings */

struct keytab
{
	short	k_code; 			/* key code */
	int 	(*k_fp)();			/* routine to handle it */
};
typedef struct keytab	KEYTAB;

/* structure for the name binding table */

struct nbind
{
	char	*n_name; 			/* name of function key */
	int 	(*n_func)();		/* function name is bound to */
};
typedef struct nbind	NBIND;

/* The editor holds deleted text chunks in the KILL buffer. The kill buffer is
 * logically a stream of ASCII characters. However, due to its unpredictable
 * size, it gets implemented as a linked list of chunks. (The d_ prefix is for
 * "deleted" text, as k_ was taken up by the keycode structure.)
 */

struct kill
{
	struct kill	*d_next; 		/* link to next chunk, NULL if last */
	char		d_chunk[KBLOCK];/* deleted text */
};
typedef struct kill 	KILL;

/* When CodEd's command interpeter needs to get a variable's name, rather than
 * it's value, it is passed back as a VDESC variable description structure.
 * The v_num field is an index into the appropriate variable table.
 */

struct vdesc
{
	int 	v_type; 			/* type of variable */
	int 	v_num;				/* ordinal pointer to variable in list */
};
typedef struct vdesc	VDESC;

/* The !WHILE directive in the execution language needs to stack references to
 * pending !WHILEs. These are stored linked to each currently open procedure
 * via a linked list of "WHBLOCK" structures.
 */

struct whblock
{
	LINE			*w_begin;	/* ptr to !while statement */
	LINE			*w_end;		/* ptr to the !endwhile statement */
	int 			w_type; 	/* block type */
	struct whblock	*w_next; 	/* next while */
};
typedef struct whblock	WHBLOCK;

#define BTWHILE 		1
#define BTBREAK 		2

#if MAGIC
/* Definitions for the metacharacters in the regular expression
 * search routines.
 */

#define MCNIL			0		/* like the '\0' for strings */
#define LITCHAR 		1		/* literal character, or string */
#define ANY 			2
#define CCL 			3
#define NCCL			4
#define BOL 			5
#define EOL 			6
#define DITTO			7
#define CLOSURE 		256 	/* an or-able value */
#define MASKCL			CLOSURE - 1

#define MC_ANY			'.' 	/* 'any' character (except NewLine) */
#define MC_BCCL			'[' 	/* begin character class */
#define MC_NCCL 		'~' 	/* negate character class */
#define MC_RCCL 		'-' 	/* range in character class */
#define MC_ECCL 		']' 	/* end of character class */
#define MC_BOL			'^' 	/* beginning of line */
#define MC_EOL			'$' 	/* end of line */
#define MC_CLOSURE		'*' 	/* closure - does not extend past newline */
#define MC_DITTO		'&' 	/* use matched string in replacement */
#define MC_ESC			'\\'	/* escape - suppress meta-meaning */

#define BIT(n)			(1 << (n))		/* an integer with one bit set */

/* HICHAR - 1 is the largest character we will deal with.
 * BMSIZE represents the number of bytes in the bitmap.
 */

#define HICHAR			256
#define BMSIZE			HICHAR >> 3

/* Typedefs that define the bitmap type for searching (BITMAP),
 * the meta-character structure for MAGIC mode searching (MC),
 * and the meta-character structure for MAGIC mode replacment (RMC).
 */

typedef char			*BITMAP;

struct mc
{
	short 	mc_type;
	union
	{
		BITMAP	cclmap;
		int 	lchar;
	} u;
};
typedef struct mc		MC;

struct rmc
{
	short	mc_type;
	char	*rstr;
};
typedef struct rmc		RMC;
#endif
